/* spice-server character device to handle a video stream

   Copyright (C) 2017 Red Hat, Inc.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, see <http://www.gnu.org/licenses/>.
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <spice/stream-device.h>

#include "char-device.h"
#include "stream-channel.h"
#include "cursor-channel.h"
#include "reds.h"

#define TYPE_STREAM_DEVICE stream_device_get_type()

#define STREAM_DEVICE(obj) \
    (G_TYPE_CHECK_INSTANCE_CAST((obj), TYPE_STREAM_DEVICE, StreamDevice))
#define STREAM_DEVICE_CLASS(klass) \
    (G_TYPE_CHECK_CLASS_CAST((klass), TYPE_STREAM_DEVICE, StreamDeviceClass))
#define STREAM_DEVICE_GET_CLASS(obj) \
    (G_TYPE_INSTANCE_GET_CLASS((obj), TYPE_STREAM_DEVICE, StreamDeviceClass))

typedef struct StreamDevice StreamDevice;
typedef struct StreamDeviceClass StreamDeviceClass;

struct StreamDevice {
    RedCharDevice parent;

    StreamDevHeader hdr;
    uint8_t hdr_pos;
    bool has_error;
    bool opened;
    bool flow_stopped;
    StreamChannel *channel;
    CursorChannel *cursor_channel;
    void *msg_buf;
    uint32_t msg_len;
};

struct StreamDeviceClass {
    RedCharDeviceClass parent_class;
};

static GType stream_device_get_type(void) G_GNUC_CONST;
static StreamDevice *stream_device_new(SpiceCharDeviceInstance *sin, RedsState *reds);

G_DEFINE_TYPE(StreamDevice, stream_device, RED_TYPE_CHAR_DEVICE)

typedef void StreamMsgHandler(StreamDevice *dev, SpiceCharDeviceInstance *sin);

static StreamMsgHandler handle_msg_format, handle_msg_data, handle_msg_invalid,
    handle_msg_cursor_set;

static RedPipeItem *
stream_device_read_msg_from_dev(RedCharDevice *self, SpiceCharDeviceInstance *sin)
{
    StreamDevice *dev = STREAM_DEVICE(self);
    SpiceCharDeviceInterface *sif;
    int n;

    if (dev->has_error || dev->flow_stopped) {
        return NULL;
    }

    sif = spice_char_device_get_interface(sin);

    /* read header */
    while (dev->hdr_pos < sizeof(dev->hdr)) {
        if (dev->msg_buf) {
            free(dev->msg_buf);
            dev->msg_buf = NULL;
            dev->msg_len = 0;
        }
        n = sif->read(sin, (uint8_t *) &dev->hdr, sizeof(dev->hdr) - dev->hdr_pos);
        if (n <= 0) {
            return NULL;
        }
        dev->hdr_pos += n;
        if (dev->hdr_pos >= sizeof(dev->hdr)) {
            dev->hdr.type = GUINT16_FROM_LE(dev->hdr.type);
            dev->hdr.size = GUINT32_FROM_LE(dev->hdr.size);
        }
    }

    switch ((StreamDevType) dev->hdr.type) {
    case STREAM_TYPE_FORMAT:
        if (dev->hdr.size != sizeof(StreamMsgFormat)) {
            handle_msg_invalid(dev, sin);
        } else {
            handle_msg_format(dev, sin);
        }
        break;
    case STREAM_TYPE_DATA:
        handle_msg_data(dev, sin);
        break;
    case STREAM_TYPE_CURSOR_SET:
        handle_msg_cursor_set(dev, sin);
        break;
    case STREAM_TYPE_CAPABILITIES:
        /* FIXME */
    default:
        handle_msg_invalid(dev, sin);
        break;
    }

    return NULL;
}

static void
handle_msg_invalid(StreamDevice *dev, SpiceCharDeviceInstance *sin)
{
    static const char error_msg[] = "Wrong protocol";

    int msg_size = sizeof(StreamMsgNotifyError) + strlen(error_msg) + 1;
    int total_size = sizeof(StreamDevHeader) + msg_size;

    RedCharDevice *char_dev = RED_CHAR_DEVICE(dev);
    RedCharDeviceWriteBuffer *buf =
        red_char_device_write_buffer_get_server_no_token(char_dev, total_size);
    buf->buf_used = total_size;

    StreamDevHeader *const hdr = (StreamDevHeader *)buf->buf;
    hdr->protocol_version = STREAM_DEVICE_PROTOCOL;
    hdr->padding = 0;
    hdr->type = GUINT16_TO_LE(STREAM_TYPE_NOTIFY_ERROR);
    hdr->size = GUINT32_TO_LE(msg_size);

    StreamMsgNotifyError *const error = (StreamMsgNotifyError *)(hdr+1);
    error->error_code = GUINT32_TO_LE(0);
    strcpy((char *) error->msg, error_msg);

    red_char_device_write_buffer_add(char_dev, buf);

    dev->has_error = true;
}

static void
handle_msg_format(StreamDevice *dev, SpiceCharDeviceInstance *sin)
{
    StreamMsgFormat fmt;
    SpiceCharDeviceInterface *sif = spice_char_device_get_interface(sin);
    int n = sif->read(sin, (uint8_t *) &fmt, sizeof(fmt));
    if (n == 0) {
        return;
    }
    if (n != sizeof(fmt)) {
        handle_msg_invalid(dev, sin);
        return;
    }
    fmt.width = GUINT32_FROM_LE(fmt.width);
    fmt.height = GUINT32_FROM_LE(fmt.height);
    stream_channel_change_format(dev->channel, &fmt);
    dev->hdr_pos = 0;
}

static void
handle_msg_data(StreamDevice *dev, SpiceCharDeviceInstance *sin)
{
    SpiceCharDeviceInterface *sif = spice_char_device_get_interface(sin);
    int n;
    while (1) {
        uint8_t buf[16 * 1024];
        n = sif->read(sin, buf, sizeof(buf));
        if (n <= 0) {
            break;
        }
        // TODO collect all message ??
        // up: we send a single frame together
        // down: guest can cause a crash
        stream_channel_send_data(dev->channel, buf, n);
        dev->hdr.size -= n;
    }
    if (dev->hdr.size == 0) {
        dev->hdr_pos = 0;
    }
}

static RedCursorCmd *
stream_msg_cursor_set_to_cursor_cmd(const StreamMsgCursorSet *msg, size_t msg_size)
{
    const QXLCursorHeader* cursor_hdr = &msg->cursor_header;

    RedCursorCmd *cmd = spice_new0(RedCursorCmd, 1);
    cmd->type = QXL_CURSOR_SET;
    cmd->u.set.position.x = 0;
    cmd->u.set.position.y = 0;
    cmd->u.set.visible = 1; // TODO
    SpiceCursor *cursor = &cmd->u.set.shape;
    cursor->header.unique = GUINT64_FROM_LE(cursor_hdr->unique);
    cursor->header.type = GUINT16_FROM_LE(cursor_hdr->type);
    cursor->header.width = GUINT16_FROM_LE(cursor_hdr->width);
    cursor->header.height = GUINT16_FROM_LE(cursor_hdr->height);
    cursor->header.hot_spot_x = GUINT16_FROM_LE(cursor_hdr->hot_spot_x);
    cursor->header.hot_spot_y = GUINT16_FROM_LE(cursor_hdr->hot_spot_y);
    if (cursor->header.width > 1024 || cursor->header.height > 1024) {
        free(cmd);
        return NULL;
    }
    size_t size_required = cursor->header.width * cursor->header.height;
    switch (cursor->header.type) {
    case SPICE_CURSOR_TYPE_ALPHA:
        size_required *= 32;
        break;
    case SPICE_CURSOR_TYPE_COLOR24:
        size_required *= 25;
        break;
    case SPICE_CURSOR_TYPE_COLOR32:
        size_required *= 33;
        break;
    default:
        free(cmd);
        return NULL;
    }
    size_required = (size_required + 7) / 8;
    if (msg_size < sizeof(StreamMsgCursorSet) + size_required) {
        free(cmd);
        return NULL;
    }
    cursor->data_size = size_required;
    cursor->data = spice_memdup(msg->data, size_required);
    return cmd;
}

static void
handle_msg_cursor_set(StreamDevice *dev, SpiceCharDeviceInstance *sin)
{
    SpiceCharDeviceInterface *sif = spice_char_device_get_interface(sin);

    // read part of the message till we get all
    if (!dev->msg_buf) {
        if (dev->hdr.size >= 1024 * 1024 * 5 || dev->hdr.size < sizeof(StreamMsgCursorSet)) {
            handle_msg_invalid(dev, sin);
            return;
        }
        dev->msg_buf = spice_malloc(dev->hdr.size);
        dev->msg_len = 0;
    }
    int n = sif->read(sin, (uint8_t *) dev->msg_buf + dev->msg_len, dev->hdr.size - dev->msg_len);
    if (n <= 0) {
        return;
    }
    dev->msg_len += n;
    if (dev->msg_len != dev->hdr.size) {
        return;
    }

    // got the full message, prepare for future message
    dev->hdr_pos = 0;

    // trasform the message to a cursor command and process it
    RedCursorCmd *cmd = stream_msg_cursor_set_to_cursor_cmd(dev->msg_buf, dev->msg_len);
    if (!cmd) {
        handle_msg_invalid(dev, sin);
        return;
    }
    cursor_channel_process_cmd(dev->cursor_channel, cmd);
}

static void
stream_device_send_msg_to_client(RedCharDevice *self, RedPipeItem *msg, RedClient *client)
{
}

static void
stream_device_send_tokens_to_client(RedCharDevice *self, RedClient *client, uint32_t tokens)
{
    spice_printerr("Not implemented!");
}

static void
stream_device_remove_client(RedCharDevice *self, RedClient *client)
{
}

static void
stream_device_stream_start(void *opaque, StreamMsgStartStop *start,
                           StreamChannel *channel G_GNUC_UNUSED)
{
    StreamDevice *dev = (StreamDevice *) opaque;

    if (!dev->opened) {
        return;
    }

    int msg_size = sizeof(*start) + sizeof(start->codecs[0]) * start->num_codecs;
    int total_size = sizeof(StreamDevHeader) + msg_size;

    RedCharDevice *char_dev = RED_CHAR_DEVICE(dev);
    RedCharDeviceWriteBuffer *buf =
        red_char_device_write_buffer_get_server_no_token(char_dev, total_size);
    buf->buf_used = total_size;

    StreamDevHeader *hdr = (StreamDevHeader *)buf->buf;
    hdr->protocol_version = STREAM_DEVICE_PROTOCOL;
    hdr->padding = 0;
    hdr->type = GUINT16_TO_LE(STREAM_TYPE_START_STOP);
    hdr->size = GUINT32_TO_LE(msg_size);

    memcpy(&hdr[1], start, msg_size);

    red_char_device_write_buffer_add(char_dev, buf);
}

static void
stream_device_stream_queue_stat(void *opaque, const StreamQueueStat *stats G_GNUC_UNUSED,
                                StreamChannel *channel G_GNUC_UNUSED)
{
    StreamDevice *dev = (StreamDevice *) opaque;

    if (!dev->opened) {
        return;
    }

    // very easy control flow... if any data stop
    // this seems a very small queue but as we use tcp
    // there's already that queue
    if (stats->num_items) {
        dev->flow_stopped = true;
        return;
    }

    if (dev->flow_stopped) {
        dev->flow_stopped = false;
        // TODO resume flow...
        // avoid recursion if we need to call get data from data handling from
        // data handling
        red_char_device_wakeup(&dev->parent);
    }
}

RedCharDevice *
stream_device_connect(RedsState *reds, SpiceCharDeviceInstance *sin)
{
    SpiceCharDeviceInterface *sif;
    ClientCbs client_cbs = { NULL, };

    StreamChannel *channel = stream_channel_new(reds);

    CursorChannel *cursor_channel = cursor_channel_new(reds, NULL, reds_get_core_interface(reds));
    client_cbs.connect = (channel_client_connect_proc) cursor_channel_connect;
    client_cbs.migrate = cursor_channel_client_migrate;
    red_channel_register_client_cbs(RED_CHANNEL(cursor_channel), &client_cbs, NULL);
    reds_register_channel(reds, RED_CHANNEL(cursor_channel));

    StreamDevice *dev = stream_device_new(sin, reds);
    dev->channel = channel;
    dev->cursor_channel = cursor_channel;
    stream_channel_register_start_cb(channel, stream_device_stream_start, dev);
    stream_channel_register_queue_stat_cb(channel, stream_device_stream_queue_stat, dev);

    sif = spice_char_device_get_interface(sin);
    if (sif->state) {
        sif->state(sin, 1);
    }

    return RED_CHAR_DEVICE(dev);
}

static void
stream_device_dispose(GObject *object)
{
    StreamDevice *device = STREAM_DEVICE(object);

    if (device->channel) {
        RedChannel *red_channel = RED_CHANNEL(device->channel);
        RedsState *reds = red_channel_get_server(red_channel);

        // prevent future connection
        reds_unregister_channel(reds, red_channel);

        // close all current connections and drop the reference
        red_channel_destroy(red_channel);
        device->channel = NULL;
    }
    if (device->cursor_channel) {
        RedChannel *red_channel = RED_CHANNEL(device->cursor_channel);
        RedsState *reds = red_channel_get_server(red_channel);

        // prevent future connection
        reds_unregister_channel(reds, red_channel);

        // close all current connections and drop the reference
        red_channel_destroy(red_channel);
        device->cursor_channel = NULL;
    }
    free(device->msg_buf);
    device->msg_buf = 0;
}

static void
stream_device_port_event(RedCharDevice *char_dev, uint8_t event)
{
    if (event != SPICE_PORT_EVENT_OPENED && event != SPICE_PORT_EVENT_CLOSED) {
        return;
    }

    StreamDevice *device = STREAM_DEVICE(char_dev);

    // reset device and channel on close/open
    device->opened = (event == SPICE_PORT_EVENT_OPENED);
    device->hdr_pos = 0;
    device->has_error = false;
    device->flow_stopped = false;
    red_char_device_reset(char_dev);
    stream_channel_reset(device->channel);
}

static void
stream_device_class_init(StreamDeviceClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS(klass);
    RedCharDeviceClass *char_dev_class = RED_CHAR_DEVICE_CLASS(klass);

    object_class->dispose = stream_device_dispose;

    char_dev_class->read_one_msg_from_device = stream_device_read_msg_from_dev;
    char_dev_class->send_msg_to_client = stream_device_send_msg_to_client;
    char_dev_class->send_tokens_to_client = stream_device_send_tokens_to_client;
    char_dev_class->remove_client = stream_device_remove_client;
    char_dev_class->port_event = stream_device_port_event;
}

static void
stream_device_init(StreamDevice *self)
{
}

static StreamDevice *
stream_device_new(SpiceCharDeviceInstance *sin, RedsState *reds)
{
    return g_object_new(TYPE_STREAM_DEVICE,
                        "sin", sin,
                        "spice-server", reds,
                        "client-tokens-interval", 0ULL,
                        "self-tokens", ~0ULL,
                        NULL);
}
